/*
wgt_vis_slinky_01

james seo
www.lossless.net

draw the current area that contains the glove
at the glove's location in the corresponding input video frame,
with the earlier areas sliding towards the left
and the later areas sliding in from the right.

this program is a visualization of data
from the white glove tracking project.
it requires input data files that are available from
www.whiteglovetracking.com
which must be placed in the proper data directory
specified by the dataPath variable.

this program will output 10059 tiff image files
if mustSaveOutputImages is set to true.
afterwards, you can save the image files as a quicktime movie
using the framerate of 29.97.

this program does not output the audio track.
you will have to add it yourself
after you have compiled the output movie.

based on wgt_viz_02_02

input data setup code and frame class
based on yellow_box_PROCESSING
by evan roth
www.ni9e.com
*/

import processing.video.*;

boolean canBeginDrawing = false;
boolean mustSaveOutputImages = false;

String appName = "wgt_vis_slinky";
String appVer = "01";
int appW = 640;
int appH = appW*3/4;
int vidW = appW;
int vidH = appH;
String dataPath = "/Users/james/Documents/code/data/";
String imgsPath = dataPath+"billie-jean-640x480/";
PImage[] imgs;
int imgsTotalNum;
PImage curImg;
Runtime r;

frame FM;
frame FMS[];
int comma;
String temp_f, temp_x, temp_y, temp_w, temp_h;
int f, x, y, w, h;
float t, inc;
Movie inMovie;
PFont courier;
String[] lines;

int curFnum;
int initFnum;
int prevFnum, nextFnum;
int minFnum = 0;
int lastGlovedFnum = 0;

int curX, curY;
int srcX, srcY, srcW, srcH;
int dstX, dstY, dstW, dstH;
int blankW = 2;

int[] maskArray;
int maxMaskVal = 240;
float maskAdj;
int dq;

/***************************************************************
* SETUP
***************************************************************/

void setup(){
  size(appW, appH);
  stroke(0);
  fill(0);
  
  courier = loadFont(dataPath+"CourierNewPSMT-14.vlw");
  textFont(courier, 16);
  
  lines = loadStrings(dataPath+"wgt_data_v1.txt");
  println("number of lines in data file: " + lines.length);

  curFnum = 0;
  
  println("setting up input data");
  setupInputData();
  println("finished setting up input data");
  println("setting up input images");
  setupInputImages();
  println("finished setting up input images");
  
  maskArray = new int[vidW*vidH];
  for (int i = 0; i < maskArray.length; i++) {
    maskArray[i] = maxMaskVal;
  }
}

void setupInputData() {
  t = 0.0f;
  inc = 1.0f / 30.0f;

  FMS = new frame[lines.length];

  for (int i=0; i < lines.length; i++) {
    comma = 0;
    temp_f = "";
    temp_x = "";
    temp_y = "";
    temp_w = "";
    temp_h = "";

    for (int j=0; j < lines[i].length(); j++) {
      char c = lines[i].charAt(j);    //this is the char we are looking at
      if(c == ','){                   //which comma are we on
        comma++;
      }
      if(comma == 0){                //break into multiple strings
        temp_f += c;
      }else if(comma == 1){
        if(c != ','){
          temp_x += c;
        }
      }else if(comma == 2){
        if(c != ','){
          temp_y += c;
        }
      }else if(comma == 3){
        if(c != ','){
          temp_h += c;
        }
      }else if(comma == 4){
        if(c != ','){
          temp_w += c;
        }
      }
    }

    f = int(temp_f);
    x = int(temp_x);
    y = int(temp_y);
    w = int(temp_w);
    h = int(temp_h);
    t += inc;

    if (x >= 640) {
      x = 639;
    }
    if (y >= 480) {
      y = 479;
    }
    if (x+w/2 >= 640) {
      w = 639-(x-w/2);
      x = 639-w/2;
    }
    if (y+h/2 >= 480) {
      h = 479-(y-h/2);
      y = 479-h/2;
    }
    
    if (x == 0 && w == 0) {
      w = blankW;
    }
    
    FM = new frame(f, x, y, w, h, t);
    FMS[i] = FM;
    //FMS[i].print();
  }
}

void setupInputImages() {
  imgsTotalNum = lines.length;
  
  imgs = new PImage[imgsTotalNum];
}

/***************************************************************
* DRAWING
***************************************************************/

void draw() {
  background(0);
  if (!canBeginDrawing) {
    initFnum = 0;
    curFnum = initFnum;
    canBeginDrawing = true;
  } else {
    println("current frame: " + curFnum);
    
    srcY = 0;
    srcH = vidH;
    dstY = srcY;
    dstH = srcH;
    
    prevFnum = curFnum;
    if (FMS[prevFnum].x > 0) {
      curX = FMS[prevFnum].x - (FMS[prevFnum].w / 2);
    } else {
      curX = FMS[lastGlovedFnum].x - (FMS[prevFnum].w / 2);
    }
    while (prevFnum >= 0) {
      if (FMS[prevFnum].x > 0) {
        srcX = FMS[prevFnum].x - (FMS[prevFnum].w / 2);
        srcW = FMS[prevFnum].w;
        if (imgs[prevFnum] == null) {
          imgs[prevFnum] = loadImage(imgsPath+"mj_bet_HD"+nf(prevFnum+1,5)+".jpg");
        }
        applyMask(prevFnum,curX);
        image(imgs[prevFnum].get(srcX,srcY,srcW,srcH),curX,srcY,srcW,srcH);
        prevFnum--;
        if (prevFnum >= 0) {
          if (curX - FMS[prevFnum].w < 0) {
            break;
          } else {
            curX -= FMS[prevFnum].w;
          }
        }
      } else {
        srcW = FMS[prevFnum].w;
        line(curX,srcY,curX,srcY+srcH);
        line(curX+1,srcY,curX+1,srcY+srcH);
        prevFnum--;
        if (prevFnum >= 0) {
          if (curX - FMS[prevFnum].w < 0) {
            break;
          } else {
            curX -= FMS[prevFnum].w;
          }
        }
      }
    }
    
    if (prevFnum < 0) {
      if (curX > 0) {
        rect(0,0,curX,srcH);
      }
    } else {
      if (curX - FMS[prevFnum].w < 0) {
        if (FMS[prevFnum].x > 0) {
          srcX = FMS[prevFnum].x - (FMS[prevFnum].w / 2) + (FMS[prevFnum].w - curX);
          srcW = curX;
          if (imgs[prevFnum] == null) {
            imgs[prevFnum] = loadImage(imgsPath+"mj_bet_HD"+nf(prevFnum+1,5)+".jpg");
          }
          if (prevFnum != curFnum) {
            applyMask(prevFnum,curX - FMS[prevFnum].w);
          }
          if (srcW > 0 && srcH > 0) {
            image(imgs[prevFnum].get(srcX,srcY,srcW,srcH),0,srcY,srcW,srcH);
          }
        } else {
          srcW = curX;
          for (int i = 0; i < srcW; i++) {
            line(i,srcY,i,srcY+srcH);
          }
        }
      }
      if (prevFnum > 0) {
        for (int i = minFnum; i < prevFnum; i++) {
          imgs[i] = null;
        }
        minFnum = prevFnum-1;
      }
    }
    
    nextFnum = curFnum;
    if (FMS[nextFnum].x > 0) {
      curX = FMS[nextFnum].x - (FMS[nextFnum].w / 2);
    } else {
      curX = FMS[lastGlovedFnum].x - (FMS[nextFnum].w / 2);
    }
    while (nextFnum < imgsTotalNum) {
      if (FMS[nextFnum].x > 0) {
        srcX = FMS[nextFnum].x - (FMS[nextFnum].w / 2);
        srcW = FMS[nextFnum].w;
        if (curX + srcW >= appW) {
          break;
        }
        if (imgs[nextFnum] == null || nextFnum == curFnum) {
          imgs[nextFnum] = loadImage(imgsPath+"mj_bet_HD"+nf(nextFnum+1,5)+".jpg");
        }
        if (nextFnum != curFnum) {
          applyMask(nextFnum,curX);
        }
        image(imgs[nextFnum].get(srcX,srcY,srcW,srcH),curX,srcY,srcW,srcH);
        curX += srcW;
        nextFnum++;
        if (nextFnum >= imgsTotalNum) {
          break;
        }
      } else {
        srcW = FMS[nextFnum].w;
        if (curX + srcW >= appW) {
          break;
        }
        line(curX,srcY,curX,srcY+srcH);
        line(curX+1,srcY,curX+1,srcY+srcH);
        curX += FMS[nextFnum].w;
        nextFnum++;
        if (nextFnum >= imgsTotalNum) {
          break;
        }
      }
    }
    
    if (nextFnum >= imgsTotalNum) {
      r = Runtime.getRuntime();
      r.gc();
      
      if (FMS[curFnum].x > 0) {
        lastGlovedFnum = curFnum;
      }
      
      curFnum++;
      if (curFnum >= imgsTotalNum) {
        println("done!");
        exit();
      }
    } else {
      if (curX + FMS[nextFnum].w >= appW) {
        if (FMS[nextFnum].x > 0) {
          srcX = FMS[nextFnum].x - (FMS[nextFnum].w / 2);
          srcW = appW-curX;
          if (imgs[nextFnum] == null || nextFnum == curFnum) {
            imgs[nextFnum] = loadImage(imgsPath+"mj_bet_HD"+nf(nextFnum+1,5)+".jpg");
          }
          if (nextFnum != curFnum) {
            applyMask(nextFnum,curX);
          }
          if (srcW > 0 && srcH > 0) {
            image(imgs[nextFnum].get(srcX,srcY,srcW,srcH),curX,srcY,srcW,srcH);
          }
        } else {
          srcW = FMS[nextFnum].w - (appW-curX);
          for (int i = 0; i < srcW; i++) {
            line(curX+i,srcY,curX+i,srcY+srcH);
          }
        }
      }
    
      r = Runtime.getRuntime();
      r.gc();
      
      if (FMS[curFnum].x > 0) {
        lastGlovedFnum = curFnum;
      }
      
      curFnum++;
      if (curFnum >= imgsTotalNum) {
        println("done!");
        exit();
      }
    }
    
    if (mustSaveOutputImages) {
      save(appName + "_" + curFnum +".tif");
    }
  }
}

/***************************************************************
* IMAGE MASK
***************************************************************/

void applyMask(int n, int x) {
  if (FMS[curFnum].x > 0) {
    dq = abs((FMS[curFnum].x-FMS[curFnum].w/2)-x);
  } else {
    dq = abs((appW/2)-1-x);
  }
  if (dq > appW) {
    dq = appW;
  }
  maskAdj = 1-((float) dq)/((float) appW*5/6);
  if (maskAdj <= 0.0001) {
    maskAdj = 0.0001;
  }
  maskAdj = tan((6*PI/32)+(9*PI/32)*maskAdj)/tan(15*PI/32);
  if (maskAdj >= 0.9999) {
    maskAdj = 0.9999;
  }
  if (FMS[curFnum].x <= 0) {
    maskAdj = 0.25;
  }
  for (int i = 0; i < maskArray.length; i++) {
    maskArray[i] = (int) (maxMaskVal*maskAdj);
  }
  imgs[n].mask(maskArray);
}